@@ -22,6 +22,7 @@
 
    Author: Raph Levien <raph@artofcode.com>
 */
+#include <string.h>
 
 #include "rsvg-shapes.h"
 #include "rsvg-styles.h"
@@ -29,6 +30,7 @@
 #include "rsvg-private.h"
 #include "rsvg-bpath-util.h"
 #include "rsvg-path.h"
+#include "rsvg-defs.h"
 
 #include <libart_lgpl/art_affine.h>
 #include <libart_lgpl/art_vpath_bpath.h>
@@ -42,6 +44,14 @@
 /* 4/3 * (1-cos 45)/sin 45 = 4/3 * sqrt(2) - 1 */
 #define RSVG_ARC_MAGIC ((double) 0.5522847498)
 
+struct _RsvgDefsPath {
+ 	RsvgDefVal super;
+	RsvgState  state;
+ 	char       *d;
+};
+ 
+typedef struct _RsvgDefsPath RsvgDefsPath;
+
 /**
  * rsvg_close_vpath: Close a vector path.
  * @src: Source vector path.
@@ -166,7 +176,7 @@ rsvg_render_bpath (RsvgHandle *ctx, const ArtBpath *bpath)
 			return;
 		}
 	
-	state = &ctx->state[ctx->n_state - 1];
+	state = rsvg_state_current (ctx);
 
 	/* todo: handle visibility stuff earlier for performance benefits 
 	 * handles all path based shapes. will handle text and images separately
@@ -260,6 +270,38 @@ rsvg_render_path(RsvgHandle *ctx, const char *d)
 	rsvg_bpath_def_free (bpath_def);
 }
 
+static void 
+rsvg_defs_path_free (RsvgDefVal *self)
+{
+	RsvgDefsPath *z = (RsvgDefsPath *)self;
+	rsvg_state_finalize (&z->state);
+	g_free (z);
+}
+
+/**
+ * Todo: Possibly publicly export this (for text)
+ */
+static void
+rsvg_handle_path (RsvgHandle *ctx, const char * d, const char * id)
+{
+	if (!ctx->in_defs)
+		rsvg_render_path (ctx, d);
+	else {
+		RsvgDefsPath *path;
+
+		if (id == NULL) 
+			return;
+
+		path = g_new (RsvgDefsPath, 1);
+		path->d = g_strdup(d);
+		rsvg_state_clone (&path->state, rsvg_state_current (ctx));
+		path->super.type = RSVG_DEF_PATH;
+		path->super.free = rsvg_defs_path_free;
+		
+		rsvg_defs_set (ctx->defs, id, &path->super);
+	}
+}
+
 void
 rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
 {
@@ -283,8 +325,8 @@ rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
 	if (d == NULL)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "path", klazz, id, atts);
-	rsvg_render_path (ctx, d);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "path", klazz, id, atts);
+	rsvg_handle_path (ctx, d, id);
 }
 
 static GString *
@@ -323,7 +365,7 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 	GString * g = NULL;
 	gchar ** pointlist = NULL;
 	const char * klazz = NULL, * id = NULL;
-	
+
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
@@ -341,7 +383,7 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 	if (!verts)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, (is_polyline ? "polyline" : "polygon"), klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), (is_polyline ? "polyline" : "polygon"), klazz, id, atts);
 	
 	/* todo: make the following more memory and CPU friendly */
 	g = rsvg_make_poly_point_list (verts);
@@ -360,9 +402,9 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 			if (!is_polyline)
 				g_string_append (d, "Z");
 			
-			rsvg_render_path (ctx, d->str);
-			g_string_free (d, TRUE);
 			g_strfreev(pointlist);
+			rsvg_handle_path (ctx, d->str, id);
+			g_string_free (d, TRUE);
 		}
 }
 
@@ -385,28 +427,33 @@ rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
 	double x1 = 0, y1 = 0, x2 = 0, y2 = 0;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
+	double font_size;
+
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "x1"))
-						x1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						x1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "y1"))
-						y1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						y1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					if (!strcmp ((char *)atts[i], "x2"))
-						x2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						x2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "y2"))
-						y2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						y2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "class"))
 						klazz = (const char *)atts[i + 1];
 					else if (!strcmp ((char *)atts[i], "id"))
 						id = (const char *)atts[i + 1];
 				}      
 		}
-	rsvg_parse_style_attrs (ctx, "line", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "line", klazz, id, atts);
 	
 	/* emulate a line using a path */
 	/* ("M %f %f L %f %f", x1, y1, x2, y2) */
@@ -420,7 +467,7 @@ rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
 	g_string_append_c (d, ' ');
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), y2));    
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -428,31 +475,36 @@ void
 rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 {
 	int i;
-	double x = -1, y = -1, w = -1, h = -1, rx = 0., ry = 0.;
+	double x = 0., y = 0., w = -1, h = -1, rx = 0., ry = 0.;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
 	gboolean got_rx = FALSE, got_ry = FALSE;
+	double font_size;
+
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "x"))
-						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "y"))
-						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "width"))
-						w = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						w = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "height"))
-						h = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						h = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "rx")) {
-						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 						got_rx = TRUE;
 					}
 					else if (!strcmp ((char *)atts[i], "ry")) {
-						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 						got_ry = TRUE;
 					}
 					else if (!strcmp ((char *)atts[i], "class"))
@@ -464,36 +516,23 @@ rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 
 	if (got_rx && !got_ry)
 		ry = rx;
-	
-	if (x < 0. || y < 0. || w < 0. || h < 0. || rx < 0. || ry < 0.)
+	else if (got_ry && !got_rx)
+		rx = ry;	
+
+	if (x < 0. || y < 0. || w <= 0. || h <= 0. || rx < 0. || ry < 0.)
 		return;
+
+	if (rx > (w / 2.))
+		rx = w / 2.;
+	if (ry > (h / 2.))
+		ry = h / 2.;
 	
-	rsvg_parse_style_attrs (ctx, "rect", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "rect", klazz, id, atts);
 	
 	/* incrementing y by 1 properly draws borders. this is a HACK */
 	y += 1.;
 	
 	/* emulate a rect using a path */
-	/*
-	d = g_strdup_printf ("M %f %f "
-						 "H %f "
-						 "A %f,%f %f,%f %f %f,%f "
-						 "V %f "
-						 "A %f,%f %f,%f %f %f,%f "
-						 "H %f "
-						 "A %f,%f %f,%f %f %f,%f "
-						 "V %f "
-						 "A %f,%f %f,%f %f %f,%f",
-						 x + rx, y,
-						 x + w - rx,
-						 rx, ry, 0., 0., 1., x + w, y + ry,
-						 y + h - ry,
-						 rx, ry, 0., 0., 1., x + w - rx, y + h,
-						 x + rx,
-						 rx, ry, 0., 0., 1., x, y + h - ry,
-						 y + ry,
-						 rx, ry, 0., 0., 1., x + rx, y);
-	*/
 
 	d = g_string_new ("M ");
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), x + rx));
@@ -572,7 +611,7 @@ rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 	g_string_append_c (d, ' ');
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), y));
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -583,21 +622,26 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 	double cx = 0, cy = 0, r = 0;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
+	double font_size;
+	
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "cx"))
-						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "cy"))
-						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "r"))
 						r = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, 
 															  rsvg_viewport_percentage((gdouble)ctx->width, (gdouble)ctx->height), 
-															  state->font_size);
+															  font_size);
 					else if (!strcmp ((char *)atts[i], "class"))
 						klazz = (const char *)atts[i + 1];
 					else if (!strcmp ((char *)atts[i], "id"))
@@ -608,23 +652,9 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 	if (cx < 0. || cy < 0. || r <= 0.)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "circle", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "circle", klazz, id, atts);
 	
 	/* approximate a circle using 4 bezier curves */
-	/* 
-	d = g_strdup_printf ("M %f %f "
-						 "C %f %f %f %f %f %f "
-						 "C %f %f %f %f %f %f "
-						 "C %f %f %f %f %f %f "
-						 "C %f %f %f %f %f %f "
-						 "Z",
-						 cx + r, cy,
-						 cx + r, cy + r * RSVG_ARC_MAGIC, cx + r * RSVG_ARC_MAGIC, cy + r, cx, cy + r,
-						 cx - r * RSVG_ARC_MAGIC, cy + r, cx - r, cy + r * RSVG_ARC_MAGIC, cx - r, cy,
-						 cx - r, cy - r * RSVG_ARC_MAGIC, cx - r * RSVG_ARC_MAGIC, cy - r, cx, cy - r,
-						 cx + r * RSVG_ARC_MAGIC, cy - r, cx + r, cy - r * RSVG_ARC_MAGIC, cx + r, cy
-						 );
-	*/
 
 	d = g_string_new ("M ");
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), cx+r));
@@ -685,7 +715,7 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 
 	g_string_append (d, " Z");
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -696,21 +726,26 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
 	double cx = 0, cy = 0, rx = 0, ry = 0;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
+	double font_size;
+	
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "cx"))
-						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "cy"))
-						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "rx"))
-						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "ry"))
-						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "class"))
 						klazz = (const char *)atts[i + 1];
 					else if (!strcmp ((char *)atts[i], "id"))
@@ -721,23 +756,10 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
 	if (cx < 0. || cy < 0. || rx <= 0. || ry <= 0.)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "ellipse", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "ellipse", klazz, id, atts);
 	
 	/* approximate an ellipse using 4 bezier curves */
-	/*
-	d = g_strdup_printf ("M %f %f "
-						 "C %f %f %f %f %f %f "
-						 "C %f %f %f %f %f %f "
-						 "C %f %f %f %f %f %f "
-						 "C %f %f %f %f %f %f "
-						 "Z",
-						 cx + rx, cy,
-						 cx + rx, cy - RSVG_ARC_MAGIC * ry, cx + RSVG_ARC_MAGIC * rx, cy - ry, cx, cy - ry,
-						 cx - RSVG_ARC_MAGIC * rx, cy - ry, cx - rx, cy - RSVG_ARC_MAGIC * ry, cx - rx, cy,
-						 cx - rx, cy + RSVG_ARC_MAGIC * ry, cx - RSVG_ARC_MAGIC * rx, cy + ry, cx, cy + ry,
-						 cx + RSVG_ARC_MAGIC * rx, cy + ry, cx + rx, cy + RSVG_ARC_MAGIC * ry, cx + rx, cy
-						 );
-	*/
+
 	d = g_string_new ("M ");
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), cx + rx));
 	g_string_append_c (d, ' ');
@@ -797,7 +819,7 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
 
 	g_string_append (d, " Z");
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -820,7 +842,12 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 	guchar *rgb = NULL;
 	int dest_rowstride;
 	double tmp_affine[6];
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
+	RsvgState *state;
+
+	/* skip over defs entries for now */
+	if (ctx->in_defs) return;
+
+	state = rsvg_state_current (ctx);
 	
 	if (atts != NULL)
 		{
@@ -847,7 +874,7 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 	if (!href || x < 0. || y < 0. || w <= 0. || h <= 0.)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "image", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, state, "image", klazz, id, atts);
 	
 	/* figure out if image is visible or not */
 	if (!state->visible)
@@ -912,3 +939,73 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 	g_object_unref (G_OBJECT (img));
 	g_free (rgb);
 }
+
+void 
+rsvg_start_use (RsvgHandle *ctx, const xmlChar **atts)
+{
+	RsvgState *state = rsvg_state_current (ctx);
+	const char * klazz = NULL, *id = NULL, *xlink_href = NULL;
+	double x = 0, y = 0, width = 0, height = 0;
+	int i;
+	double affine[6];
+	gboolean got_width = FALSE, got_height = FALSE;
+
+	if (atts != NULL)
+		{
+			for (i = 0; atts[i] != NULL; i += 2)
+				{
+					if (!strcmp ((char *)atts[i], "x"))
+						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+					else if (!strcmp ((char *)atts[i], "y"))
+						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+					else if (!strcmp ((char *)atts[i], "width")) {
+						width = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						got_width = TRUE;
+					}
+					else if (!strcmp ((char *)atts[i], "height")) {
+						height = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);					
+						got_height = TRUE;
+					}
+					else if (!strcmp ((char *)atts[i], "class"))
+						klazz = (const char *)atts[i + 1];
+					else if (!strcmp ((char *)atts[i], "id"))
+						id = (const char *)atts[i + 1];
+					else if (!strcmp ((char *)atts[i], "xlink:href"))
+						xlink_href = (const char *)atts[i + 1];
+				}
+		}
+	
+	/* < 0 is an error, 0 disables rendering. TODO: handle positive values correctly */
+	if (got_width || got_height)
+		if (width <= 0. || height <= 0.)
+			return;
+	
+	if (xlink_href != NULL)
+		{
+			RsvgDefVal * parent = rsvg_defs_lookup (ctx->defs, xlink_href+1);
+			if (parent != NULL)
+				switch(parent->type)
+					{
+					case RSVG_DEF_PATH:
+						{
+							RsvgDefsPath *path = (RsvgDefsPath*)parent;
+
+							/* combine state definitions */
+							rsvg_state_clone (state, &path->state);
+							art_affine_translate (affine, x, y);
+							art_affine_multiply (state->affine, affine, state->affine);
+
+							rsvg_parse_style_attrs (ctx, state, "use", klazz, id, atts);
+							if (state->opacity != 0xff)
+								rsvg_push_opacity_group (ctx);
+
+							/* always want to render inside of a <use/> */
+							rsvg_render_path (ctx, path->d);
+							break;
+						}
+					default:
+						g_warning ("Unhandled defs entry/type %s %d\n", id, parent->type);
+						return;
+					}
+		}
+}
